package com.mparticle.smartype.generator

import com.github.ajalt.clikt.core.CliktCommand
import com.github.ajalt.clikt.core.subcommands
import com.github.ajalt.clikt.parameters.options.default
import com.github.ajalt.clikt.parameters.options.option
import com.github.ajalt.clikt.parameters.options.prompt
import com.github.ajalt.clikt.parameters.options.required
import com.github.ajalt.clikt.parameters.types.choice
import com.github.ajalt.clikt.parameters.types.file
import com.mparticle.smartype.generator.adapters.DefaultAdapterFactory
import com.mparticle.smartype.generator.adapters.MParticleDataPlanAdapter
import kotlinx.serialization.json.Json
import kotlinx.serialization.json.JsonConfiguration
import kotlinx.serialization.json.JsonObject
import java.io.File
import java.io.IOException
import java.util.*


class Init : CliktCommand(name="init", help = "Initialize a 'smartype.config.json' configuration file") {
    val ios by option().choice("yes", "no").prompt("Would you like to generate iOS?", default = "no")
    val android by option().choice("yes", "no").prompt("Would you like to generate Android?", default = "no")
    val binaryOutputDirectory by option().prompt("Where should Smartype libraries be generated", default = "smartype-dist")
    val apiSchemaFile by option().prompt("Please specify a file path containing the JSON schema that you'd like to use for generation")

    override fun run() {

        var options = GeneratorOptions(
            IOSOptions(ios == "yes"),
            AndroidOptions(android == "yes"),
            binaryOutputDirectory,
            apiSchemaFile
        )
        var jsonOptions = JsonConfiguration(
            encodeDefaults = true,
            ignoreUnknownKeys = false,
            isLenient = false,
            serializeSpecialFloatingPointValues = false,
            allowStructuredMapKeys = true,
            prettyPrint = true,
            unquotedPrint = false,
            useArrayPolymorphism = false
        )
        val json = Json(jsonOptions)
        val jsonData = json.stringify(GeneratorOptions.serializer(), options)
        File("smartype.config.json").writeText(jsonData)
        echo("smartype.config.json generated successfully! Have fun :-)")

    }
}

class Generate : CliktCommand(name="generate", help = "Generate Smartype Client libraries given a Smartype configuration file") {
    private val smartypeConfigFile by option(help="A configuration file generated by the init command")
        .file().default(File("smartype.config.json"))
    private val GRADLEW_EXECUTABLE = "gradlew"
    private val TEMP_DIR = ".smartype/"
    private fun jsonContents(file: File): JsonObject {
        return Json.parseJson(file.readText()).jsonObject
    }

    private fun runningInJar() : Boolean {
        val protocol = this.javaClass.getResource("").protocol
        return Objects.equals(protocol, "jar")
    }

    private fun extractJar(targetDirectory: String) {
        val jarFile = File(
                Generate::class.java.getProtectionDomain().getCodeSource().getLocation()
                        .toURI()
        ).getPath()
        val pb = ProcessBuilder(listOf("unzip","-o", "-q", jarFile, "-d", targetDirectory, "-x", "*.class"))
        pb.redirectOutput(ProcessBuilder.Redirect.INHERIT)
        pb.redirectError(ProcessBuilder.Redirect.INHERIT)
        val p = pb.start()
        p.waitFor()
    }

    override fun run() {
        val json = Json(JsonConfiguration.Stable)
        val options = json.parse(GeneratorOptions.serializer(), smartypeConfigFile.readText())

        val inJar = runningInJar()

        if (inJar) {
            extractJar(TEMP_DIR)
        }

        val jsonSchema = jsonContents(File(options.apiSchemaFile))
        //TODO: enable additional adapters based on configuration
        val adapter = DefaultAdapterFactory().createFromName(MParticleDataPlanAdapter().getName())
        val analyticsSchema = adapter.extractSchemas(jsonSchema)
        val smartTypeClass = SmartypeObject()
        smartTypeClass.configureApi(analyticsSchema)

        var outDirectory = TEMP_DIR + "smartype-generator/build/generatedSources"
        if (!inJar) {
            outDirectory = "build/generatedSources"
        }
        smartTypeClass.finalize(outDirectory)

        try {
            var gradleBinDir = TEMP_DIR
            var projectDirectory = TEMP_DIR
            var binOutputDirectory = options.binaryOutputDirectory
            if (!inJar) {
                gradleBinDir = "../"
                projectDirectory = "../"
                binOutputDirectory = File(binOutputDirectory).absolutePath
            } else {
                val parentDirectory = File(
                    Generate::class.java.getProtectionDomain().getCodeSource().getLocation()
                        .toURI()
                ).parentFile
                if (!File(binOutputDirectory).isAbsolute) {
                    binOutputDirectory = parentDirectory.resolve(
                        binOutputDirectory
                    ).absolutePath
                }
            }

            val gradleArgs = mutableListOf<String>()
            if (projectDirectory.isNotBlank()) {
                gradleArgs.add("-p")
                gradleArgs.add(projectDirectory)
            }

            if (options.androidOptions.enabled) {
                gradleArgs.add(":smartype:bundleReleaseAar")
            }
            if (options.iosOptions.enabled) {
                gradleArgs.add(":smartype:iosFatFramework")
            }

            val gradleCommand = listOf(gradleBinDir + GRADLEW_EXECUTABLE) + gradleArgs
            val pb2 = ProcessBuilder(gradleCommand)
            pb2.redirectOutput(ProcessBuilder.Redirect.INHERIT)
            pb2.redirectError(ProcessBuilder.Redirect.INHERIT)
            val p2 = pb2.start()
            p2.waitFor()

            val directory =
                File(binOutputDirectory)
            directory.deleteRecursively()
            if (!directory.exists()) {
                directory.mkdirs()
            }

            val iosBuildDirectory = File(projectDirectory).resolve("smartype/build/ios")
            if (iosBuildDirectory.exists()) {
                val mviOS =
                    listOf("mv", iosBuildDirectory.absolutePath, File(binOutputDirectory).absolutePath + "/")
                val pb3 = ProcessBuilder(mviOS)
                pb3.redirectOutput(ProcessBuilder.Redirect.INHERIT)
                pb3.redirectError(ProcessBuilder.Redirect.INHERIT)
                val p3 = pb3.start()
                p3.waitFor()
            } else if (options.iosOptions.enabled){
                println("Unable to locate built iOS binaries in ${iosBuildDirectory.absolutePath}")
            }

            val androidBuildDirectory = File(projectDirectory).resolve("smartype/build/outputs/aar")
            if (androidBuildDirectory.exists()) {
                val mvAndroid =
                    listOf("mv", androidBuildDirectory.absolutePath, File(binOutputDirectory).resolve("android/").absolutePath)
                val pb4 = ProcessBuilder(mvAndroid)
                pb4.redirectOutput(ProcessBuilder.Redirect.INHERIT)
                pb4.redirectError(ProcessBuilder.Redirect.INHERIT)
                val p4 = pb4.start()
                p4.waitFor()
            } else if (options.androidOptions.enabled){
                println("Unable to locate built Android binaries in ${androidBuildDirectory.absolutePath}")
            }

        } catch (e: IOException) {
            e.printStackTrace()
        }
    }
}

class Generator: CliktCommand(printHelpOnEmptyArgs = true) {
    override fun run() = Unit
}

fun main(args: Array<String>) = Generator().subcommands(Generate(), Init()).main(args)
